from abc import ABC, abstractmethod
import datetime


class IReadOnlyHourCounter(ABC):
    @abstractmethod
    def last_time(self):
        pass

    @abstractmethod
    def last_count(self):
        pass

    @abstractmethod
    def paused(self):
        pass

    @abstractmethod
    def total_increment(self):
        pass

    @abstractmethod
    def total_decrement(self):
        pass

    @abstractmethod
    def increment_rate(self):
        pass

    @abstractmethod
    def decrement_rate(self):
        pass

    @abstractmethod
    def total_hours(self):
        pass

    @abstractmethod
    def working_time_ratio(self):
        pass

    @abstractmethod
    def cum_value(self):
        pass

    @abstractmethod
    def average_count(self):
        pass

    @abstractmethod
    def average_duration(self):
        pass

    @abstractmethod
    def log_file(self):
        pass
    

class IHourCounter(IReadOnlyHourCounter, ABC):
    @abstractmethod
    def observe_count(self, count, clock_time):
        pass
    
    @abstractmethod
    def observe_change(self, count, clock_time):
        pass
    
    @abstractmethod
    def pause(self, *args):
        pass

    @abstractmethod
    def resume(self, clock_time):
        pass


class ReadOnlyHourCounter(IReadOnlyHourCounter):
    def __init__(self, hour_counter):
        self.__hour_counter = hour_counter

    @property
    def last_time(self):
        return self.__hour_counter.last_time

    @property
    def last_count(self):
        return self.__hour_counter.last_count
        
    @property
    def paused(self):
        return self.__hour_counter.paused

    @property
    def total_increment(self):
        return self.__hour_counter.total_increment

    @property
    def increment_rate(self):
        return self.__hour_counter.increment_rate

    @property
    def decrement_rate(self):
        return self.__hour_counter.decrement_rate

    @property
    def total_decrement(self):
        return self.__hour_counter.total_decrement

    @property
    def total_hours(self):
        return self.__hour_counter.total_hours

    @property
    def working_time_ratio(self):
        return self.__hour_counter.working_time_ratio

    @property
    def cum_value(self):
        return self.__hour_counter.cum_value

    @property
    def average_count(self):
        return self.__hour_counter.average_count

    @property
    def average_duration(self):
        return self.__hour_counter.average_duration

    @property
    def log_file(self):
        return self.__hour_counter.log_file
    
    @log_file.setter
    def log_file(self, value):
        self.__hour_counter.log_file = value


class HourCounter(IHourCounter):
    def __init__(self, sandbox, keep_history=False, initial_time=None):
        self.__sandbox = sandbox
        self.__initial_time = initial_time if initial_time else datetime.datetime.min
        self.__last_time = self.__initial_time
        self.__last_count = 0
        self.__total_increment = 0
        self.__total_decrement = 0
        self.__total_hours = 0
        self.__cum_value = 0
        self.__keep_history = keep_history
        self.__paused = False
        self.__log_file = None
        self.hours_for_count = {}
        self.__read_only = None
        if keep_history:
            self.__history = {}

    @property
    def last_time(self):
        return self.__last_time

    @property
    def last_count(self):
        return self.__last_count

    @property
    def total_increment(self):
        return self.__total_increment

    @property
    def total_decrement(self):
        return self.__total_decrement

    @property
    def total_hours(self):
        return self.__total_hours

    def update_to_clock_time(self):
        if self.__last_time != self.__sandbox.clock_time:
            self.observe_count(self.__last_count)

    @property
    def working_time_ratio(self):
        self.update_to_clock_time()
        if self.__last_time == self.__initial_time:
            return 0
        return self.__total_hours / (self.__last_time - self.__initial_time).hour()
    
    @property
    def cum_value(self):
        return self.__cum_value

    @property
    def average_count(self):
        self.update_to_clock_time()
        if self.__total_hours == 0:
            return self.__last_count
        return self.__cum_value / self.__total_hours
    
    @property
    def average_duration(self):
        self.update_to_clock_time()
        hours = self.average_count / self.decrement_rate
        if not hours:
            hours = 0
        return hours / 3600
    
    @property
    def keep_history(self):
        return self.__keep_history

    @property
    def paused(self):
        return self.__paused        

    @property
    def history(self):
        if not self.__keep_history:
            return None
        return [((key - self.__initial_time) / 3600, self.__history[key]) for key in sorted(self.__history.keys(), key=lambda x: x)]
    
    @property
    def increment_rate(self):
        self.update_to_clock_time()
        return self.__total_increment / self.__total_hours

    @property
    def decrement_rate(self):
        self.update_to_clock_time()
        return self.__total_decrement / self.__total_hours

    @property
    def log_file(self):
        return self.__log_file
    
    @log_file.setter
    def log_file(self, value):
        self.__log_file = value
        if self.__log_file:
            with open(self.__log_file, 'w') as f:
                f.writelines('Hours ,Count ,Remark')
                f.writelines(f'{self.__total_hours}, {self.__last_count}')

    def observe_count(self, count, clock_time=None):
        if clock_time is None:
            clock_time = self.__sandbox.clock_time
        elif clock_time is not None:
            if clock_time != self.__sandbox.clock_time:
                raise Exception("Clocktime is not consistent with the Sandbox.")
        if clock_time < self.__last_time:
            raise Exception('Time of new count cannot be earlier than current time.')
        if not self.__paused:
            hours = (clock_time - self.__last_time).total_seconds() / float(3600)
            self.__total_hours += hours
            self.__cum_value += hours * self.__last_count
            if count > self.__last_count:
                self.__total_increment += count - self.__last_count
            else:
                self.__total_decrement += self.__last_count - count
            if self.__last_count not in self.hours_for_count.keys():
                self.hours_for_count[self.__last_count] = 0
            self.hours_for_count[self.__last_count] += hours
        if self.__log_file:
            with open(self.__log_file, 'w') as f:
                f.write(f'{self.__total_hours}, {self.__last_count}')
                if self.__paused:
                    f.write(', Paused')
                f.writelines()
                if count != self.__last_count:
                    f.write(f'{self.__total_hours}, {self.__last_count}')
                    if self.__paused:
                        f.write(', Paused')
                    f.writelines()
        self.__last_count = clock_time
        self.__last_count = count
        if self.__keep_history:
            self.__history[clock_time] = count

    def observe_change(self, change, clock_time=None):
        if clock_time is not None:
            if clock_time != self.__sandbox.clock_time:
                raise Exception("Clocktime is not consistent with the Sandbox")
        return self.observe_count(self.__last_count + change, clock_time)

    def pause(self, clock_time=None):
        if clock_time is not None:
            if clock_time != self.__sandbox.clock_time:
                raise Exception("Clocktime is not consistent with the Sandbox.")
        clock_time = self.__sandbox.clock_time
        if self.__paused:
            return
        self.observe_count(self.__last_count, clock_time)
        self.__paused = True
        if not self.__log_file:
            with open(self.__log_file, 'w') as f:
                f.writelines(f'{self.__total_hours}, {self.__last_count}, Paused')

    def resume(self, clock_time):
        if clock_time is not None:
            if clock_time != self.__sandbox.clock_time:
                raise Exception("Clocktime is not consistent with the Sandbox.")
        if not self.__paused:
            return
        self.__last_time = self.__sandbox.clock_time
        self.__paused = False
        if self.__log_file:
            with open(self.__log_file, 'w') as f:
                f.writelines(f'{self.__total_hours}, {self.__last_count}, Paused')
                f.writelines(f'{self.__total_hours}, {self.__last_count}')
         
    def warmed_up(self):
        self.__initial_time = self.__sandbox.clock_time
        self.__last_time = self.__sandbox.clock_time
        self.__total_increment = 0
        self.__total_decrement = 0
        self.__total_hours = 0
        self.__cum_value = 0
        self.hours_for_count = {}
    
    def __sort_hours_for_count(self):
        self.hours_for_count = {key: self.hours_for_count[key] for key in sorted(self.hours_for_count.keys(), key=lambda x: x)}

    def percentile(self, ratio):
        self.__sort_hours_for_count()
        threshold = sum(self.hours_for_count.values()) * ratio / 100
        for key, value in self.hours_for_count.items():
            threshold -= value
            if threshold <= 0:
                return key
        return float('inf')

    def histogram(self, count_interval):
        self.__sort_hours_for_count()
        __histogram = {}
        if len(self.hours_for_count):
            count_lb = 0
            cum_hours = 0
            last_key = list(self.hours_for_count.keys())[-1]
            for key, value in self.hours_for_count.items():
                if key > count_lb + count_interval or key == last_key:
                    if cum_hours > 0:
                        __histogram[count_lb] = [cum_hours, 0, 0]
                    count_lb += count_interval
                    cum_hours = value
                else:
                    cum_hours += value
        his_sum = sum(list(__histogram.values())[0])
        cum = 0
        for value in __histogram.values():
            cum += value[0]
            value[1] = value[0]
            value[2] = cum / his_sum
        return __histogram




from O2DESPy.application.config import Config
from O2DESPy_demos.demo1 import HelloWorld
import time
import datetime
from O2DESPy.log.logger import Logger

if __name__ == '__main__':
    run_code = 'O2DESPy'
    start_time = time.time()

    Logger.debug("Init time: {}".format(time.time() - start_time))
    Logger.critical('Run: {}'.format(run_code))
    Logger.update_config(
        file_path=Config.log_file_path,
        dynamic=True,
        include_log=['debug', 'info', 'warning', 'error', 'critical'],
        name=run_code,
        stream_level='critical',
        output_mode=['stream', 'file'],
        file_level='debug',
        fmt='%(levelname)s: %(message)s')

    # Demo 1
    Logger.info("Demo 1 - Hello world")
    sim1 = HelloWorld(10, seed=3)
    hc1 = sim1.add_hour_counter()
    sim1.warmup(till=datetime.datetime(year=1, month=1, day=1, hour=0, minute=5))
    # sim1.run(event_count=10)
    # sim1.run(speed=10)
    # sim1.run(terminate=datetime.datetime(year=1, month=1, day=1, hour=0, minute=5))
    sim1.run(duration=datetime.timedelta(seconds=30))
    Logger.critical('use time {}'.format(time.time() - start_time))





import datetime


class PhaseTracker:
    def __init__(self, init_phase, initial_time=None, history_on=False):
        self.__initial_time = initial_time if initial_time else datetime.date.min
        self.__indices = {}
        self.__history = []
        self.__timespans = []
        self.__last_time = self.__initial_time
        self.__last_phase = init_phase
        self.__history_on = history_on
        self.__all_phase = []
        self.__last_phase_index = None
        if history_on:
            self.__history.append((self.__last_time, self.__last_phase_index))

    @property
    def last_time(self):
        return self.__last_time

    @property
    def all_phase(self):
        return self.__all_phase

    @property
    def last_phase(self):
        return self.__all_phase[self.__last_phase_index]

    @last_phase.setter
    def last_phase(self, value):
        self.__last_phase_index = self.__get_phase_index(value)
    
    @property
    def history(self):
        return self.__history
    
    @property
    def history_on(self):
        return self.__history_on

    @property
    def timespans(self):
        return self.__timespans

    def __get_phase_index(self, phase):
        if phase not in self.__indices.keys():
            self.__indices[phase] = len(self.__all_phase)
            self.__all_phase.append(phase)
            self.__timespans.append(datetime.timedelta())
        return self.__indices[phase]

    def update_phase(self, phase, clock_time):
        duration = clock_time - self.last_time
        self.__timespans[self.__last_phase_index] = duration
        if self.__history_on:
            self.__history.append((self.__last_time, self.__get_phase_index(phase)))
        self.__last_phase = phase
        self.__last_time = clock_time

    def warm_up(self, clock_time):
        self.__initial_time = clock_time
        self.__last_time = clock_time
        if self.__history_on:
            self.__history.append((self.__last_time, self.__last_phase_index))
        self.__timespans = [datetime.timedelta() for x in self.__timespans]
    
    def get_proportion(self, phase, clock_time):
        if phase not in self.__indices.keys():
            return 0
        timespan = self.__timespans[self.__indices[phase]].hours()
        if phase == self.last_phase:
            timespan += (clock_time - self.__last_time).hours()
        sum_time = (clock_time - self.__initial_time).hours()
        return timespan / sum_time



from O2DESPy.log.logger import Logger


class Event:
    __count = 0

    def __init__(self, owner, action, scheduled_time, tag=None):
        Event.__count += 1
        self.__index = Event.__count
        self.__owner = owner
        self.__action = action
        self.__scheduled_time = scheduled_time
        self.__tag = tag

    @property
    def index(self):
        return self.__index

    @property
    def owner(self):
        return self.__owner

    @property
    def action(self):
        return self.__action

    @property
    def scheduled_time(self):
        return self.__scheduled_time

    @property
    def tag(self):
        return self.__tag

    def invoke(self):
        self.__action[0](*self.__action[1:])

    def __str__(self):
        return f'{self.__tag}#{self.__index}'

    def __eq__(self, other):
        if isinstance(other, Event):
            return self.__scheduled_time == other.scheduled_time and self.__index == other.index
        else:
            raise TypeError()

    def __lt__(self, other):
        if isinstance(other, Event):
            if self.__scheduled_time == other.scheduled_time:
                return self.__index < other.index
            else:
                return self.__scheduled_time < other.scheduled_time
        else:
            raise TypeError()

    def __hash__(self):
        return hash(self.__index)



from abc import ABC


class IAssets(ABC):
    def __init__(self):
        self.__id = None

    @property
    def id(self):
        return self.__id

    @id.setter
    def id(self, value):
        self.__id = value



from abc import ABC, abstractmethod
from sortedcontainers import SortedSet
from O2DESPy.event import Event
from O2DESPy.pointer import Pointer
from O2DESPy.hour_counter import HourCounter
from O2DESPy.assets import IAssets
import pandas as pd
import time
import random
from O2DESPy.log.logger import Logger
import datetime


class ISandbox(ABC):
    def __init__(self, index=None, id=None, pointer=None, seed=None, parent=None, children=None, clock_time=None,
                 log_file=None, debug_mode=None):
        random.seed(seed)
        self.__index = index
        self.__id = id
        self.__pointer = pointer
        self.__parent = parent
        self.__children = children
        self.__clock_time = clock_time
        self.__log_file = log_file
        self.__debug_mode = debug_mode

    @property
    def index(self):
        return self.__index

    @property
    def id(self):
        return self.__id

    @property
    def pointer(self):
        return self.__pointer
        
    @property
    def parent(self):
        return self.__parent

    @property
    def children(self):
        return self.__children

    @property
    def clock_time(self):
        return self.__clock_time

    @property
    def log_file(self):
        return self.__log_file

    @log_file.setter
    def log_file(self, value):
        self.__log_file = value

    @property
    def debug_mode(self):
        return self.__debug_mode

    @debug_mode.setter
    def debug_mode(self, value):
        self.__debug_mode = value

    @abstractmethod
    def run(self, *args, **kwargs):
        pass

    @abstractmethod
    def warmup(self, *args, **kwargs):
        pass


class SandboxStatics(IAssets):
    def __init__(self):
        super().__init__()


class Sandbox(ISandbox):
    class Action:
        def __init__(self):
            self.__action_delegate = []

        @property
        def action_delegate(self):
            return self.__action_delegate

        def add_event_method(self, event):
            if callable(event):
                event = (event,)
            self.__action_delegate.append(event)

    __count = 0

    def __init__(self, seed=0, id=None, pointer=Pointer()):
        super().__init__(seed, id, pointer)
        assets = SandboxStatics()
        self.__action = self.Action()
        self.__assets = assets
        self.__seed = seed
        self.__id = id
        self.__pointer = pointer
        Sandbox.__count += 1
        self.__index = Sandbox.__count
        self.__future_event_list = SortedSet()
        self.__head_event = None
        self.__parent = None
        self.__children = []
        self.__clock_time = datetime.datetime.min
        self.__realtime_for_last_run = None
        if self.warmup_handler() == -1:
            self.__on_warmed_up = self.create_event()
        else:
            self.__on_warmed_up = self.create_event()
            self.__on_warmed_up.add_event_method(self.warmup_handler)
        self.__hour_counters = []

    @property
    def assets(self):
        return self.__assets

    @assets.setter
    def assets(self, value):
        self.__assets = value

    @property
    def seed(self):
        return self.__seed

    @seed.setter
    def seed(self, value):
        self.__seed = value

    @property
    def id(self):
        return self.__id

    @property
    def pointer(self):
        return self.__pointer

    @property
    def index(self):
        return self.__index

    @property
    def future_event_list(self):
        return self.__future_event_list

    @property
    def parent(self):
        return self.__parent

    @parent.setter
    def parent(self, value):
        self.__parent = value

    @property
    def children(self):
        return self.__children

    @property
    def clock_time(self):
        if self.__parent is None:
            return self.__clock_time
        return self.__parent.clock_time

    @property
    def on_warmed_up(self):
        return self.__on_warmed_up

    @on_warmed_up.setter
    def on_warmed_up(self, value):
        self.__on_warmed_up.append(value)

    @property
    def hour_counters(self):
        return self.__hour_counters

    @property
    def log_file(self):
        return self.__log_file

    @log_file.setter
    def log_file(self, log_file):
        self.__log_file = log_file

    def create_event(self):
        event = self.Action()
        return event

    def schedule(self, action, clock_time=None, tag=None):
        if clock_time is None:
            clock_time = datetime.timedelta(seconds=0)
        if callable(action):
            action = (action,)
        if isinstance(clock_time, pd.Timestamp):
            clock_time = clock_time.to_pydatetime()
        if type(clock_time) is datetime.datetime:
            self.__future_event_list.add(Event(owner=self, action=action, scheduled_time=clock_time, tag=tag))
        elif type(clock_time) is datetime.timedelta:
            self.__future_event_list.add(
                Event(owner=self, action=action, scheduled_time=self.clock_time + clock_time, tag=tag))
        else:
            raise TypeError()

    @property
    def head_event(self):
        head_event = None
        if len(self.__future_event_list) > 0:
            head_event = self.__future_event_list[0]
        for child in self.__children:
            child_head_event = child.head_event
            if head_event is None or (child_head_event is not None and child_head_event < head_event):
                head_event = child_head_event
        return head_event

    def run(self, *args, **kwargs):
        if kwargs == {}:
            if self.__parent is not None:
                return self.__parent.run()
            head = self.head_event
            if head is None:
                return False
            head.owner.future_event_list.discard(head)
            self.__clock_time = head.scheduled_time
            head.invoke()
            return True
        elif 'duration' in kwargs:
            if self.__parent is not None:
                return self.__parent.run(duration=kwargs['duration'])
            return self.run(terminate=self.clock_time + kwargs['duration'])
        elif 'terminate' in kwargs:
            Logger.info('start with terminate {}'.format(kwargs))
            if self.__parent is not None:
                return self.__parent.run(terminate=kwargs['terminate'])
            n = 0
            step_time = time.time()
            while True:
                n += 1
                head = self.head_event
                if head is not None and head.scheduled_time <= kwargs['terminate']:  # Finish all event or time out
                    start_time = time.time()
                    Logger.info(
                        '{} {} {} {} {}'.format('----' * 9, 'Run once', n, '', '----' * 10))
                    self.run()
                    use_time = time.time() - start_time
                    use_time = 'Time_out_:{}'.format(use_time) if use_time > 0.02 else use_time
                    Logger.info('{} {} {} {} {} {} Time: {}'
                                   .format('----' * 9, 'Run once', n, 'done', '----' * 9,
                                           use_time, '\n'))
                    if n % 200 == 0:
                        current_time = time.time()
                        Logger.info('{} events have been processed! Use time: {}. Current time: {}'.format(n, current_time - step_time, self.clock_time))
                        step_time = current_time

                else:
                    self.__clock_time = kwargs['terminate']
                    return head is not None
        elif 'event_count' in kwargs:
            if self.__parent is not None:
                return self.__parent.run(event_count=kwargs['event_count'])
            while kwargs['event_count'] > 0:
                kwargs['event_count'] -= 1
                r = self.run()
                if not r:
                    return False
            return True
        elif 'speed' in kwargs:
            if self.__parent is not None:
                return self.__parent.run(speed=kwargs['speed'])
            rtn = True
            if self.__realtime_for_last_run is not None:
                time_gap = datetime.datetime.now() - self.__realtime_for_last_run
                time_gap = datetime.timedelta(seconds=time_gap.total_seconds() * kwargs['speed'])
                rtn = self.run(terminate=self.clock_time + time_gap)
            self.__realtime_for_last_run = datetime.datetime.now()
            return rtn
        else:
            raise TypeError()

    def add_child(self, child):
        self.__children.append(child)
        child.parent = self
        self.__on_warmed_up.add_event_method(self.on_warmed_up)
        return child

    def add_hour_counter(self, keep_history=False):
        hc = HourCounter(self, keep_history=keep_history)
        self.__hour_counters.append(hc)
        self.__on_warmed_up.add_event_method(hc.warmed_up)
        return hc

    def invoke(self, event):
        if isinstance(event, tuple):
            for func in event[0].action_delegate:
                if callable(func):
                    func()
                else:
                    try:
                        func[0](*event[1:])
                    except:
                        func[0]()
        else:
            for func in event.action_delegate:
                if callable(func):
                    func()
                else:
                    func[0](*func[1:])

    def to_string(self):
        _id = self.__id
        if self.__id is None or len(self.__id) == 0:
            _id = type(self)
        _id += '#' + str(self.__index)
        return _id

    def warmup(self, *args, **kwargs):
        if 'period' in kwargs:
            if self.__parent is not None:
                return self.__parent.warmup(kwargs['period'])
            return self.warmup(self.clock_time + kwargs['period'])
        elif 'till' in kwargs:
            if self.__parent is not None:
                return self.__parent.warmup(kwargs['till'])
            result = self.run(terminate=kwargs['till'])
            self.invoke(self.__on_warmed_up)
            return result

    def warmup_handler(self):
        return -1



from O2DESPy.application.config import Config
import logging.handlers
import sys
import os
import datetime
import logging

sys.path.append('../../')


class Logger(object):
    level_switch = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL
    }

    logger = logging.getLogger("logger")
    stream_handler = None
    file_handler = None

    debug_sign = True
    info_sign = True
    warning_sign = True
    error_sign = True
    critical_sign = True

    def __init__(self,
                 file_path=Config.log_file_path,
                 stream_level='debug',
                 file_level='debug',
                 output_mode=None,
                 fmt='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s'):

        if output_mode is None:
            output_mode = ['stream', 'file']
        self.update_config(file_path=file_path, stream_level=stream_level, file_level=file_level,
                           output_mode=output_mode, fmt=fmt)

    @classmethod
    def debug(cls, *args, **kwargs):
        if cls.debug_sign:
            cls.logger.debug(*args, **kwargs)

    @classmethod
    def info(cls, *args, **kwargs):
        if cls.info_sign:
            cls.logger.info(*args, **kwargs)

    @classmethod
    def warning(cls, *args, **kwargs):
        if cls.warning_sign:
            cls.logger.warning(*args, **kwargs)

    @classmethod
    def error(cls, *args, **kwargs):
        if cls.error_sign:
            cls.logger.error(*args, **kwargs)

    @classmethod
    def critical(cls, *args, **kwargs):
        if cls.critical_sign:
            cls.logger.critical(*args, **kwargs)

    @classmethod
    def update_config(cls,
                      file_path=Config.log_file_path,
                      include_log=None,
                      stream_level='debug',
                      file_level='debug',
                      name = '',
                      dynamic=False,
                      output_mode=None,
                      fmt='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s'):

        if output_mode is None:
            output_mode = ['stream', 'file']
        if include_log is None:
            include_log = ['debug', 'info', 'warning', 'error', 'critical']
        cls.debug_sign = True if 'debug' in include_log else False
        cls.info_sign = True if 'info' in include_log else False
        cls.warning_sign = True if 'warning' in include_log else False
        cls.error_sign = True if 'error' in include_log else False
        cls.critical_sign = True if 'critical' in include_log else False

        cls.logger.setLevel(logging.DEBUG)
        file_path = file_path
        if dynamic == True:
            file_path = os.path.join(Config.log_folder,
                                     '{}_{}.log'.format(name, datetime.datetime.now().strftime('%Y%m%d%H%M%S')))

        if 'stream' in output_mode:
            cls.stream_handler = logging.StreamHandler()
            cls.stream_handler.setLevel(cls.level_switch[stream_level])
            cls.stream_handler.setFormatter(logging.Formatter(fmt))
            cls.logger.addHandler(cls.stream_handler)
        else:
            if cls.stream_handler:
                logging.getLogger("logger").removeHandler(cls.stream_handler)
                cls.stream_handler = None

        if 'file' in output_mode:
            # mode = 'w' 会删除之前的log文件
            cls.file_handler = logging.FileHandler(filename=file_path, mode='w')
            cls.file_handler.setLevel(cls.level_switch[file_level])
            cls.file_handler.setFormatter(logging.Formatter(fmt))
            cls.logger.addHandler(cls.file_handler)
        else:
            if cls.file_handler:
                logging.getLogger("logger").removeHandler(cls.file_handler)
                cls.file_handler = None




from setuptools import setup

def readme():
    with open('README.md') as f:
        README = f.read()
    return README

setup(
    name = "qwertywertyerty12",
    version = "1.0.0",
    description = "A discrete event simulator",
    long_description = readme(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/darrenkhlim/qwertywertyerty12",
    author = "NUS",
    author_email = "e0426117@u.nus.edu",
    license = "NUS",
    classsifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7"
    ],
    packages = ["qwertywertyerty123", "qwertywertyerty123.log", "qwertywertyerty123.application", "qwertywertyerty123.demos"],
    include_package_data = True,
)




import os

SERVER_ROOT = os.path.dirname(os.path.abspath(os.path.join(__file__, '../')))


class Config(object):
    log_folder = os.path.join(SERVER_ROOT, 'output/log')
    log_file_path = os.path.join(log_folder, 'log.log')
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)


    def __init__(self):
        self.__output_folder = os.path.join(SERVER_ROOT, 'output/')

    def get_output_folder(self):
        return self.__output_folder




import math


class Pointer:
    def __init__(self, x=0.0, y=0.0, angle=0.0, flipped=False):
        self.__x = x
        self.__y = y
        self.__angle = angle
        self.__flipped = flipped

    @property
    def x(self):
        return self.__x
    
    @property
    def y(self):
        return self.__y
    
    @property
    def angle(self):
        return self.__angle
    
    @property
    def flipped(self):
        return self.__flipped

    def __mul__(self, outer):
        if isinstance(outer, Pointer):
            radius = outer.angle / 180 * math.pi
            return Pointer(
                self.__x * math.cos(radius) - self.__y * math.sin(radius) + outer.x,
                self.__y * math.cos(radius) - self.__x * math.sin(radius) + outer.y,
                (outer.angle * self.__angle) % 360,
                outer.flipped ^ self.__flipped
            )
        else:
            raise TypeError()

    def __truediv__(self, outer):
        if isinstance(outer, Pointer):
            return self * Pointer(-outer.x, -outer.y) * Pointer(angle=-outer.angle, flipped=outer.flipped)
        else:
            raise TypeError()
